/*
 * Buffer API. In namespace p_buf_. 
 *
 * Copyright (c) 2005, Adam Gleave
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the 
 *     documentation and/or other materials provided with the distribution.
 *  3. Neither the name of the copyright holders nor the names of its
 *     contributors may be used to endorse or promote products derived from
 *     this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT OWNER AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

/* 
 * @UNSAFE: WHOLE FILE
 */

#include <stdio.h>
#include <stdlib.h>
#include <stddef.h>
#include <string.h>

#include "buffer.h"
#include "bufdang.h"
#include "privfail.h"

static int p_buf_private_alloc(p_arr *, size_t);
static int p_buf_private_check_read(const p_arr *, size_t, size_t);
static int p_buf_private_check_write(p_arr *,size_t, size_t);

#define ROK P_RET_OK
#define RFAIL P_RET_FAIL

#define RSAME P_BUF_RET_SAME
#define RDIFF P_BUF_RET_DIFF

/* PURPOSE
 * To cause the function to return paramater ret, unless p_buf has option
 * hard enabled; in which case the program terminates with errmsg.
 * PARAMATERS
 * errmsg       - Error message to report.
 * ret          - Value to return. */

#define FAIL_WITH_RET(errmsg, ret)                                         \
  if (p_buf->hard == 1)                                                    \
    p_f_die(errmsg);                                                       \
  return ret

/* PURPOSE
 * To cause the function to return RFAIL, unless p_buf has option hard
 * enabled; in which case the program terminates with errmsg.
 * PARAMATERS
 * errmsg       - Error message to report. */
 
#define FAIL(errmsg) FAIL_WITH_RET(errmsg, RFAIL)

/* +NOTE: p_arr is declared in p_arr.h. */

/* Prviate Functions */

/* PURPOSE
 * To allocate space for a buffer.
 * PARAMATERS
 * p_buf        - Pointer to p_arr object.
 * p_size       - Size to allocate for buffer.
 * RETURN VALUE
 * ROK on success, RFAIL on failure. */

static int p_buf_private_alloc(p_arr *p_buf, size_t p_size)
{
  unsigned char *tmp_array;
  
  /* Are we attempting to allocate for a read only buffer? */
  if (p_buf->readonly == 1) {
    p_f_bug("p_buf_private_alloc: allocation attempted for a readonly buffer."); 
  }
  /* No? Good */
  else { /* TAG: else1. */
    /* Is the request for allocation already what we have? */
    if (p_buf->alloced == p_size)
      return ROK; 
    
    /* Would allocation allocated more than the limit? */
    if (p_size > p_buf->limit) {
      FAIL("p_buf_private_alloc: allocation past limit.");
    }
    
    /* Reallocate, storing (possibly moved) address in tmp_array. */
    if ((tmp_array = realloc((void *)p_buf->array, p_size)) == NULL) {
      /* EEEK, realloc failed. */ 
      FAIL("p_buf_private_alloc: realloc failed.");
    } else { /* realloc succeeded - correct p_buf->array */
      p_buf->array = tmp_array;
      p_buf->alloced = p_size;
    }
  } /* END of TAG: else1. */
  
  return ROK;
}

/* PURPOSE
 * To check that p_pos is allowed to be read from. 
 * PARAMATERS
 * p_buf        - Constant pointer to p_arr object.
 * p_pos        - Pointer to position in array. 
 * p_datasize   - Data size wishing to be read, in bytes.
 * RETURN VALUES
 * ROK for allowed, RFAIL for disallowed. */

static int p_buf_private_check_read(const p_arr *p_buf, size_t p_pos,
                                    size_t p_datasize)
{
  /* 
   * +SECNOTE: It is possible that p_datasize is larger than p_buf->used.
   * In this case, p_buf->used-p_datasize would result in a negative value
   * and, as size_t is unsigned, a very large positive value! This would
   * result in a possible buffer overflow as the check would be circumvented.
   * This is why p_datasize is checked to be <= to p_buf->used.
   */
  if (p_datasize > p_buf->used) {
    FAIL("p_buf_private_check_read: p_datasize is larger than used amount "
         "amount of buffer.");
  }
  
  /* Is maximium position (including datasize) past USED amount buffer? */
  if (p_pos > p_buf->used-p_datasize) {
    FAIL("p_buf_private_check_read: read would go past used buffer.");
  }
  
  return ROK;
}

/* PURPOSE
 * To check that p_pos is allowed to be written to. 
 * PARAMATERS
 * p_buf        - Pointer to p_arr structure.
 * p_pos        - Pointer to position in array. 
 * p_datasize   - Data size wishing to be written, in bytes.
 * RETURN VALUES
 * ROK for allowed, RFAIL for disallowed. */

static int p_buf_private_check_write(p_arr *p_buf, size_t p_pos,
                                     size_t p_datasize)
{
  size_t size;
  
  /* Is buffer readonly? */
  if (p_buf->readonly == 1) {
    FAIL("p_buf_private_check_write: write to a readonly buffer.");
  }
  
  /* size=maximium size p_datasize can be without overflowing size_t. */
  size = (size_t)-1 - p_pos;
  
  /* Does p_datasize exceed maximium allowed size? */
  if (p_datasize > size) {
    FAIL("p_buf_private_check_write: write exceeds size_t.");
  }
  /* p_datasize+p_pos won't overflow size_t - great! */
  else { /* TAG: ELSE1 */
    /* size=write to check */
    size = p_pos+p_datasize;
    /* Is size not within ALLOCATED amount of buffer? */
    if (size > p_buf->alloced) {
      /* +FIXME: Need a more effecient way of allocating, reallocating
       * each time growth is needed is very ineffecient. */
      if (p_buf_private_alloc(p_buf, size) != ROK) {
        return RFAIL;
      }
    }
  } /* END: ELSE1 */
  
  /* Is the write exceeding used amount of buffer? */
  if (size > p_buf->used)
    p_buf->used = size;
  
  return ROK;
}

/* Public Functions - Initializing Functions */

/* @SECNOTICE: We use malloc to allocate our p_arr structures,
 * not calloc. Because of this, we have non-zeroed data. So, we
 * must explicitly overwrite all data.
 */

void p_buf_init_static(p_arr *p_buf, size_t p_size)
{
  /* +SECNOTE: Unitialized variables in p_buf: array, limit, alloced, used, 
   * readonly, hard. */
  /* Does allocating space for buffer fail? */
  if ((p_buf->array = malloc(p_size)) == NULL) 
    p_f_die("p_buf_init_static: unable to allocate space for buffer.");
  
  /* +SECNOTE: Unitialized variables in p_buf: limit, alloced, used,
   * readonly, hard. */
  p_buf->limit = p_buf->alloced = p_size;
  
  /* +SECNOTE: Unitialized variables in p_buf: used, readonly, hard. */
  p_buf->used = p_buf->readonly = 0;
  
  /* +SECNOTE: Unitialized variables in p_buf: hard. */
  p_buf-> hard = 1;
  
  /* +SECNOTE: All variables in p_buf initialized. */
}

int p_buf_init_static_soft(p_arr *p_buf, size_t p_size)
{
  /* +SECNOTE: Unitialized variables in p_buf: array, limit, alloced, used, 
   * readonly, hard. */
  /* Does allocating space for buffer fail? */
  if ((p_buf->array = malloc(p_size)) == NULL)
		return RFAIL;
  
  /* +SECNOTE: Unitialized variables in p_buf: limit, alloced, used,
   * readonly, hard. */
  p_buf->limit = p_buf->alloced = p_size;
  
  /* +SECNOTE: Unitialized variables in p_buf: used, readonly, hard. */
  p_buf->used = p_buf->readonly = 0;
  
  /* +SECNOTE: Unitialized variables in p_buf: hard. */
  p_buf-> hard = 0;
  
  /* +SECNOTE: All variables in p_buf initialized. */
  
  return ROK;
}

void p_buf_init_dynamic(p_arr *p_buf, size_t p_isize, size_t p_msize)
{
  /* +SECNOTE: Unitialized variables in p_buf: array, limit, alloced, used, 
   * readonly, hard. */
  /* Does allocating space for buffer fail? */
  if ((p_buf->array = malloc(p_isize)) == NULL)
    p_f_die("p_buf_init_dynamic: can't allocate space for buffer.");
  
  /* +SECNOTE: Unitialized variables in p_buf: limit, alloced, used,
   * readonly, hard. */
  p_buf->limit = p_msize;
  
  /* +SECNOTE: Unitialized variables in p_buf: alloced, used, readonly,
   * hard. */
  p_buf->alloced = p_isize;
  
  /* +SECNOTE: Unitialized variables in p_buf: used, readonly, hard. */
  p_buf->used = p_buf->readonly = 0;
  
  /* +SECNOTE: Unitialized variables in p_buf: hard. */
  p_buf->hard = 1;
  
  /* +SECNOTE: All variables in p_buf initialized. */
}

int p_buf_init_dynamic_soft(p_arr *p_buf, size_t p_isize, size_t p_msize)
{
  /* +SECNOTE: Unitialized variables in p_buf: array, limit, alloced, used, 
   * readonly, hard. */
  /* Does allocating space for buffer fail? */
  if ((p_buf->array = malloc(p_isize)) == NULL)
    return RFAIL;
  
  /* +SECNOTE: Unitialized variables in p_buf: limit, alloced, used,
   * readonly, hard. */
  p_buf->limit = p_msize;
  
  /* +SECNOTE: Unitialized variables in p_buf: alloced, used, readonly,
   * hard. */
  p_buf->alloced = p_isize;
  
  /* +SECNOTE: Unitialized variables in p_buf: used, readonly, hard. */
  p_buf->used = p_buf->readonly = 0;
  
  /* +SECNOTE: Unitialized variables in p_buf: hard. */
  p_buf->hard = 0;
  
  /* +SECNOTE: All variables in p_buf initialized. */
  
  return ROK;
}

/* Public Functions - Allocating Functions */

p_arr *p_buf_create_static(size_t p_size)
{
  p_arr *buf;
  
  /* Does allocating space for p_arr structure fail? */
  if ((buf = malloc(sizeof(p_arr))) == NULL)
    p_f_die("p_buf_create_static: can't allocate space for buffer.");
  
  p_buf_init_static(buf, p_size);
  
  return buf;
}

p_arr *p_buf_create_static_soft(size_t p_size)
{
  p_arr *buf;
  
  /* Does allocating space for p_arr structure fail? */
  if ((buf = malloc(sizeof(p_arr))) == NULL) 
    return NULL;
  
  if (p_buf_init_static_soft(buf, p_size) != ROK) {
    free(buf);
    return NULL;
  }
  
  return buf;
}

p_arr *p_buf_create_dynamic(size_t p_isize, size_t p_msize)
{
  p_arr *buf;
  
  /* Does allocating space for p_arr structure fail? */
  if ((buf = malloc(sizeof(p_arr))) == NULL)
    p_f_die("p_buf_create_dynamic: can't allocate space for buffer.");
  
  p_buf_init_dynamic(buf, p_isize, p_msize);
  
  return buf;
}

p_arr *p_buf_create_dynamic_soft (size_t p_isize, size_t p_msize)
{
  p_arr *buf;
  
  /* Does allocating space for p_arr structure fail? */
  if ((buf = malloc(sizeof(p_arr))) == NULL)
    return NULL;
  
  if (p_buf_init_dynamic_soft(buf, p_isize, p_msize) != ROK) {
    free(buf);
    return NULL;
  }
  
  return buf;
}

/* Public Functions - Deallocating */

void p_buf_free(p_arr *p_buf)
{
  /* Buffer has been allocated? */
  if (p_buf->alloced == 1)
    free((void *)p_buf->array);
  free(p_buf);
}

void p_buf_freebuf(p_arr *p_buf)
{
  /* Buffer has been allocated? */
  if (p_buf->alloced == 1)
    free((void *)p_buf->array);
}


/* Public Functions - Option Setters */
/* +NOTE: Try and keep these functions down to a minimium, this is meant to be
 * a small API after all :) */

void p_buf_set_readonly(p_arr *p_buf)
{
  p_buf->readonly = 1;
}

/* Public Functions - Information Retrievers */
/* + NOTE: Try and keep these functions down to a minimium, this meant to be
 * a small API after all :) */

size_t p_buf_get_used(p_arr *p_buf)
{
  return p_buf->used;
}

/* Public Functions - Editing Functions */
int p_buf_write(p_arr *p_buf, size_t p_pos, const void *p_data,
                size_t p_datasize)
{
  /* Are we not allowed to write this amount to p_buf? */
  if (p_buf_private_check_write(p_buf, p_pos, p_datasize) != ROK)
    return RFAIL;
  
  memcpy(p_buf->array + p_pos, p_data, p_datasize);
  
  return ROK;
}

int p_buf_append_data(p_arr *p_buf, const void *p_data, size_t p_datasize)
{
  return p_buf_write(p_buf, p_buf->used, p_data, p_datasize);
}

int p_buf_insert_data(p_arr *p_buf, size_t p_pos, const void *p_data,
                size_t p_datasize)
{
  size_t size, move_size;
  
  /* move_size = used amount of buffer. */ 
  move_size = p_buf->used;
  /* Does position exceed used amount of buffer? (remember, position starts at
   * 0 - so position 0 would mean that used would have to be 1 to be ok.) */
  if (p_pos >= move_size) {
    FAIL("p_buf_insert_data: cannot insert past used amount of buffer.");
  }
  /* move_size = amount of bytes to move away. */
  move_size -= p_pos;
  
  /* Would move_size + p_datasize overflow size_t? */
  if (p_datasize > (size_t)-1 - move_size) {
    FAIL("p_buf_insert_data: number of bytes to modify after p_pos would "
         "overflow size_t.");
  }
  
  /* size = number of bytes to modify after pos. */
  size = p_datasize + move_size;
  
  if (p_buf_private_check_write(p_buf, p_pos, size) != ROK)
    return RFAIL;
	
  memmove(p_buf->array + p_pos + p_datasize, p_buf->array + p_pos,
          move_size);
  memcpy(p_buf->array + p_pos, p_data, p_datasize);
  
  return ROK;
}

int p_buf_insert(p_arr *p_dst_buf, size_t p_dst_pos,
                 const p_arr *p_src_buf, size_t p_src_pos, size_t p_len)
{
  size_t size, move_size;
  
  /* move_size = used amount of destiation buffer. */ 
  move_size = p_dst_buf->used;
  /* Does destination position exceed used amount of destination buffer? */
  if (p_dst_pos >= move_size) {
    if (p_dst_buf->hard == 1)
      p_f_die("p_buf_insert: cannot insert past used amount of buffer.");
    return RFAIL;
  }
  
  /* move_size = amount of bytes to move away in destination buffer. */
  move_size -= p_dst_pos;
  
  /* Would move_size + p_len overflow size_t? */
  if (p_len > (size_t)-1 - move_size) {
    if (p_dst_buf->hard == 1)
      p_f_die("p_buf_insert: number of bytes to modify after p_pos "
              "would overflow size_t.");
    return RFAIL;
  }
  
  /* size = number of bytes to modify after p_dst_pos in destination. */
  size = p_len + move_size;

  if (p_buf_private_check_read(p_src_buf, p_src_pos, p_len) != ROK)
    return RFAIL;
  
  if (p_buf_private_check_write(p_dst_buf, p_dst_pos, size) != ROK)
    return RFAIL;
	
  memmove(p_dst_buf->array + p_dst_pos + p_len,
          p_dst_buf->array + p_dst_pos, move_size);
  memcpy(p_dst_buf->array + p_dst_pos, p_src_buf->array + p_src_pos, p_len);
  
  return ROK;
}

int p_buf_insert_buf(p_arr *p_dst_buf, size_t p_dst_pos,
                     const p_arr *p_src_buf)
{
  return p_buf_insert(p_dst_buf, p_dst_pos, p_src_buf, 0, p_src_buf->used);
}

int p_buf_copy(p_arr *p_dst_buf, size_t p_dst_pos,
               const p_arr *p_src_buf, size_t p_src_pos, size_t p_len)
{
  if (p_buf_private_check_read(p_src_buf, p_src_pos, p_len) != ROK)
    return RFAIL;
  
  if (p_buf_private_check_write(p_dst_buf, p_dst_pos, p_len) != ROK)
    return RFAIL;
  
  /* Do buffers overlap? */
  if (p_dst_buf == p_src_buf) {
    memmove(p_dst_buf->array + p_dst_pos,
            p_src_buf->array + p_src_pos, p_len);
  } else {
    memcpy(p_dst_buf->array + p_dst_pos,
           p_src_buf->array + p_src_pos, p_len);
	}
  
  return ROK;
}

int p_buf_append(p_arr *p_dst_buf, const p_arr *p_src_buf,
                 size_t p_src_pos, size_t p_len)
{
  return p_buf_copy(p_dst_buf, p_dst_buf->used, p_src_buf, p_src_pos, p_len);
}

int p_buf_append_buf(p_arr *p_dst_buf, const p_arr *p_src_buf)
{
  return p_buf_append(p_dst_buf, p_src_buf, 0, p_src_buf->used);
}

int p_buf_duplicate(p_arr *p_dst_buf, const p_arr *p_src_buf)
{
  if (p_buf_copy(p_dst_buf, 0, p_src_buf, 0, p_src_buf->used) != ROK)
    return RFAIL;
  
  p_buf_truncate(p_dst_buf, p_src_buf->used);
  
  return ROK;
}

int p_buf_delete(p_arr *p_buf, size_t p_pos, size_t p_size)
{
  size_t rem_size;
  
  if (p_buf->readonly == 1) {
    FAIL("p_buf_delete: attempt to delete in a readonly buffer.");
  }
  
  /* rem_size = used amount of buffer */
  rem_size = p_buf->used;
  
  /* Does position exceed used amount of buffer? */
  if (p_pos >= rem_size) {
    FAIL("p_buf_delete: position past used buffer.");
  }
  
  /* rem_size = maximum possible to delete after p_pos */
  rem_size -= p_pos;
  
  if (p_size < rem_size) {
    /* We delete from between. */ 
    rem_size -= p_size;
    memmove(p_buf->array + p_pos, p_buf->array + p_pos + p_size, rem_size);
  } else if (p_size > rem_size) {
    /* AIEE! p_size exceeds rem_size. */
    FAIL("p_buf_delete: would delete past used buffer.");
  } else {
    /* Truncate the buffer. */
    rem_size = 0;
  }
  
  p_buf->used = p_pos + rem_size;
  
  return ROK;
}

void p_buf_empty(p_arr *p_buf)
{
  p_buf->used = 0;
}

/* Public Functions - Editing Functions - DANGEROUS */

/*
 * @WARNING: These function(s) are baad things. Don't use them. Ever.
 */

int p_buf_truncate(p_arr *p_buf, size_t p_size)
{
  /* Would we be making the used amount larger? */
  if (p_size > p_buf->used) {
    FAIL("p_buf_truncate: attempt to extend used buffer.");
  }
  
  p_buf->used = p_size;
  
  return ROK;
}

int p_buf_grow(p_arr *p_buf, size_t p_size)
{
  if (p_buf_private_check_write(p_buf, p_buf->used, p_size) != ROK)
    return RFAIL; 
	
  return ROK;
}

/* Public Functions - Reading Functions */

const void *p_buf_read(p_arr *p_buf)
{
  if (p_buf_private_check_read(p_buf, 0, 1) != ROK) 
    return NULL;
  
  return p_buf->array;
}

const void *p_buf_read_pos(p_arr *p_buf, size_t p_pos, size_t p_size)
{
  if (p_buf_private_check_read(p_buf, p_pos, p_size) != ROK)
    return NULL;
  
  return p_buf->array + p_pos;
}

/* Public Functions - Reading Functions - DANGEROUS */

/*
 * @WARNING: These function(s) are baad things. Don't use them. Ever.
 */

void *p_buf_read_modifiable(p_arr *p_buf)
{
  if (p_buf_private_check_write(p_buf, 0, 1) != ROK)
    return NULL;
  
  return p_buf->array;
}

void *p_buf_read_pos_modifiable(p_arr *p_buf, size_t p_pos, size_t p_size)
{
  if (p_buf_private_check_write(p_buf, p_pos, p_size) != ROK)
    return NULL;
  
  return p_buf->array + p_pos;
}

/* Public Functions - Operations */

int p_buf_cmp(p_arr *p_buf1, p_arr *p_buf2)
{
  if (p_buf1->used != p_buf2->used)
    return RDIFF;
  
  if (memcmp(p_buf1->array, p_buf2->array, p_buf1->used) != 0)
    return RDIFF;
  
  return RSAME;
}
