/*
 * Copyright (c) 2005, Adam Gleave
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the 
 *     documentation and/or other materials provided with the distribution.
 *  3. Neither the name of the copyright holders nor the names of its
 *     contributors may be used to endorse or promote products derived from
 *     this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT OWNER AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

#ifndef P_BUFFER_H
#define P_BUFFER_H

#include "arr.h"
#include "returns.h"

#define P_BUF_RET_SAME 0
#define P_BUF_RET_DIFF 1

/* PURPOSE
 * To initialize a p_arr structure to static size of p_size.
 * PARAMATERS
 * p_buf        - Pointer to p_arr structure.
 * p_size		    - Size of array.
 * RETURN VALUES
 * Any failure will kill the program. */ 

void p_buf_init_static(p_arr *p_buf, size_t p_size);

/* PURPOSE
 * To initialize a p_arr structure to static size of p_size,
 * with soft checking.
 * PARAMATERS
 * p_buf        - Pointer to p_arr structure.
 * p_size       - Size of array.
 * RETURN VALUES
 * Returns P_RET_OK on success, P_RET_FAIL on failure. */

int p_buf_init_static_soft(p_arr *p_buf, size_t p_size);

/* PURPOSE
 * To initialize a p_arr structure to a dynamic size of p_isize
 * with a maximium size of p_msize.
 * PARAMATERS
 * p_buf        - Pointer to p_arr structure.
 * p_isize      - Initial size of array.
 * p_msize      - Maximium size of array.
 * RETURN VALUES
 * Any failure will kill the program.
 * NOTES
 * If no limit is wanted, (size_t)-1 can be used for p_msize. */

void p_buf_init_dynamic(p_arr *p_buf, size_t p_isize, size_t p_msize);

/* PURPOSE
 * To initialize a p_arr structure with soft checking to a
 * dynamic size of p_isize with a maximium size of p_msize.
 * PARAMATERS
 * p_buf        - Pointer to p_arr structure.
 * p_isize      - Initial size of array.
 * p_msize      - Maximium size of array.
 * RETURN VALUES
 * Returns P_RET_OK on success, P_RET_FAIL on failure. 
 * NOTES
 * If no limit is wanted, (size_t)-1 can be used for p_msize. */

int p_buf_init_dynamic_soft(p_arr *p_buf, size_t p_isize, size_t p_msize);

/* PURPOSE
 * To create a p_arr scruture and initialize it to
 * have a static size of p_size.
 * PARAMATERS
 * p_size       - Size of array.
 * RETURN VALUES
 * Returns pointer to newly created p_arr structure.
 * On failure, it kills the program. */

p_arr *p_buf_create_static(size_t p_size);

/* PURPOSE
 * To create a p_arr structure and initialize it to
 * have a static size of p_size, with soft checking.
 * PARAMATERS
 * p_size       - Size of array.
 * RETURN VALUES
 * Returns pointer to newly created p_arr structure on success,
 * NULL on failure. */

p_arr *p_buf_create_static_soft(size_t p_size);

/* PURPOSE
 * To create a p_arr structure and initialize it to have 
 * a dynamic size of p_isize with a maximium size of p_msize.
 * PARAMATERS
 * p_isize      - Initial size of array.
 * p_msize      - Maximium size of array.
 * RETURN VALUES
 * Returns pointer to newly created p_arr structure.
 * On failure, it kills the program. */

p_arr *p_buf_create_dynamic(size_t p_isze, size_t p_msize);

/* PURPOSE
 * To create a p_arr structure with soft checking and initialize it
 * to have a dynamic size of p_isize with a maximium size of p_msize.
 * PARAMATERS
 * p_isize      - Initial size of array.
 * p_msize      - Maximium size of array.
 * RETURN VALUES
 * Returns pointer to newly created p_arr structure on success,
 * NULL on failure. */

p_arr *p_buf_create_dynamic_soft(size_t p_isize, size_t p_msize);

/* PURPOSE
 * To free, if there is one, p_buf's buffer and p_buf itself.
 * PARAMATERS
 * p_buf        - Pointer to p_arr structure. */

void p_buf_free(p_arr *p_buf);

/* PURPOSE
 * To free, if there is one, p_buf's buffer.
 * PARAMATERS
 * p_buf        - Pointer to buffer structure. */

void p_buf_freebuf(p_arr *p_buf); 

/* PURPOSE
 * To mark a buffer readonly.
 * PARAMATERS
 * p_buf        - Pointer to p_arr structure. */

void p_buf_set_readonly(p_arr *p_buf);

/* PURPOSE
 * To retrieve the used number of bytes.
 * PARAMATERS
 * p_buf        - Pointer to p_arr structure.
 * RETURN VALUES
 * The used number of bytes of p_buf. */

size_t p_buf_get_used(p_arr *p_buf);

/* PURPOSE
 * To write p_datasize bytes of data to p_buf starting at p_pos.
 * PARAMATERS
 * p_buf        - Pointer to p_arr structure.
 * p_pos        - Position in p_buf to start writing at.
 * p_data       - Pointer to data to be written.
 * p_datasize   - Amount of data in bytes to be written.
 * RETURN VALUES
 * Returns P_RET_OK on success, P_RET_FAIL on failure. */

int p_buf_write(p_arr *p_buf, size_t p_pos, const void *p_data,
				size_t p_datasize);

/* PURPOSE
 * To append p_datasize bytes of data to p_buf.
 * PARAMATERS
 * p_buf        - Pointer to p_arr structure.
 * p_data       - Pointer to data to be written.
 * p_datasize   - Amount of data in bytes to be written.
 * RETURN VALUES
 * Returns P_RET_OK on success, P_RET_FAIL on failure. */

int p_buf_append_data(p_arr *p_buf, const void *p_data, size_t p_datasize);

/* PURPOSE
 * To insert p_datasize bytes of data into p_buf at p_pos.
 * PARAMATERS
 * p_buf        - Pointer to p_arr structure.
 * p_pos        - Position in p_buf to insert data at.	
 * p_data       - Pointer to data to be written.
 * p_datasize   - Amount of data in bytes to be written.
 * RETURN VALUES
 * Returns P_RET_OK on success, P_RET_FAIL on failure. */

int p_buf_insert_data(p_arr *p_buf, size_t p_pos, const void *p_data,
                      size_t p_datasize);

/* PURPOSE
 * To insert p_len bytes of data from p_src_buf starting at
 * p_src_buf into p_dst_buf starting at p_dst_pos.
 * PARAMATERS
 * p_dst_buf    - Pointer to p_arr structure of destination.
 * p_dst_pos    - Position in p_dst_buf to insert data at.
 * p_src_buf    - Pointer to p_arr structure of source.
 * p_src_pos    - Position in p_src_buf to start inserting from.
 * p_len        - Amount of data in bytes to be written.
 * RETURN VALUES
 * Returns P_RET_OK on success, P_RET_FAIL on failure. */

int p_buf_insert(p_arr *p_dst_buf, size_t p_dst_pos,
                 const p_arr *p_src_buf, size_t p_src_pos, size_t p_len);

/* PURPOSE
 * To insert p_src_buf into p_dst_buf at position p_dst_pos.
 * PARAMATERS
 * p_dst_buf    - Pointer to p_arr structure of destination.
 * p_dst_pos    - Position in p_dst_buf to insert data at.
 * p_src_buf    - Pointer to p_arr structure of source.
 * RETURN VALUES
 * Returns P_RET_OK on success, P_RET_FAIL on failure. */

int p_buf_insert_buf(p_arr *p_dst_buf, size_t p_dst_pos,
                     const p_arr *p_src_buf);
/* PURPOSE
 * To copy p_len bytes from p_src_buf starting at p_src_pos
 * into p_dst_buf at p_dst_pos.
 * PARAMATERS
 * p_dst_buf    - Pointer to p_arr structure of destination.
 * p_dst_pos    - Position in p_dst_buf to start copying to.
 * p_src_buf    - Pointer to p_arr structure of source.
 * p_src_pos    - Position in p_src_buf to start copying from.
 * p_len        - Number of bytes to copy.
 * RETURN VALUES
 * Returns P_RET_OK on success, P_RET_FAIL on failure. */

int p_buf_copy(p_arr *p_dst_buf, size_t p_dst_pos,
               const p_arr *p_src_buf, size_t p_src_pos, size_t p_len);

/* PURPOSE
 * To append p_len bytes from p_src_buf starting at p_src_pos to p_dst_buf.
 * PARAMATERS
 * p_dst_buf    - Pointer to p_arr structure of destination.
 * p_src_buf    - Pointer to p_arr structure of source.
 * p_src_pos    - Position in p_src_buf to starting copying from.
 * p_len        - Number of bytes to append.
 * RETURN VALUES
 * Returns P_RET_OK on success, P_RET_FAIL on failure. */

int p_buf_append(p_arr *p_dst_buf, const p_arr *p_src_buf,
                 size_t p_src_pos, size_t p_len);

/* PURPOSE
 * To append p_src_buf to p_dst_buf.
 * PARAMATERS
 * p_dst_buf    - Pointer to p_arr structure of destination.
 * p_src_buf    - Pointer to p_arr structure of source.
 * RETURN VALUES
 * Returns P_RET_OK on success, P_RET_FAIL on failure. */

int p_buf_append_buf(p_arr *p_dst_buf, const p_arr *p_src_buf);

/* PURPOSE
 * To make p_dst_buf a duplicate of p_src_buf in data (i.e. not settings.)
 * PARAMATERS
 * p_dst_buf    - Pointer to p_arr structure of destination.
 * p_src_buf    - Pointer to p_arr structure of source.
 * RETURN VALUES
 * Returns P_RET_OK on success, P_RET_FAIL on failure. */

int p_buf_duplicate(p_arr *p_dst_buf, const p_arr *p_src_buf);

/* PURPOSE
 * To delete p_size bytes of data from p_buf starting at p_pos.
 * PARAMATERS
 * p_buf        - Pointer to p_arr structure.
 * p_pos        - Position in p_buf to start deleting data from.
 * p_size       - Amount of data in bytes to be deleted.
 * RETURN VALUES
 * Returns P_RET_OK on success, P_RET_FAIL on failure. */

int p_buf_delete(p_arr *p_buf, size_t p_pos, size_t p_size);

/* PURPOSE
 * To delete all data in p_buf.
 * PARAMATERS
 * p_buf        - Pointer to p_arr structure.
 * NOTES
 * You should ask yourself why your using this function - would freeing
 * and then creating a new buffer make more sense? */

void p_buf_empty(p_arr *p_buf);

/* PURPOSE
 * To retrieve the buffer belonging to p_buf.
 * PARAMATERS
 * p_buf        - Pointer to p_arr structure.
 * RETURN VALUES
 * Returns p_buf's buffer on success, NULL on failure. */

const void *p_buf_read(p_arr *p_buf);

/* PURPOSE
 * To retrieve the buffer belonging to p_buf at position p_pos
 * checking that buffer is large enough to read p_size bytes.
 * PARAMATERS
 * p_buf        - Pointer to p_arr structure.
 * p_pos        - Position in p_buf to read from.
 * p_size       - Number of bytes to read.
 * RETURN VALUES
 * Returns p_buf's buffer with p_pos added to it on success, NULL on failure. */

const void *p_buf_read_pos(p_arr *p_buf, size_t p_pos, size_t p_size);

/* PURPOSE
 * To check if p_buf1's buffer is equal to p_buf2's buffer.
 * PARAMATERS
 * p_buf1       - Pointer to first p_arr structure.
 * p_buf2       - Pointer to second p_arr structure.
 * RETURN VALUES
 * Returns P_BUF_RET_SAME if buffers are equal, P_BUF_RET_DIFF if they are different. */

int p_buf_cmp(p_arr *p_buf1, p_arr *p_buf2);

#endif /* P_BUFFER_H */
