/*
 * POSIX IO API. Uses buffer API. In name space p_io_.   
 *
 * Copyright (c) 2005, Adam Gleave
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the 
 *     documentation and/or other materials provided with the distribution.
 *  3. Neither the name of the copyright holders nor the names of its
 *     contributors may be used to endorse or promote products derived from
 *     this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT OWNER AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

/* 
 * @UNSAFE: Whole file.
 */

#include <stdio.h>
#include <sys/types.h>
#include <unistd.h>

#include "buffer.h"
#include "bufdang.h"
#include "iounix.h"
#include "priviounixnet.h"

#define ROK P_RET_OK

#define RIOPINTFAIL P_IOP_RET_INTFAIL

ssize_t p_io_read(p_arr *p_buf, int p_fd, size_t p_size)
{
  ssize_t ret;
  size_t used;
  
  used = p_buf_get_used(p_buf);
  
  if (p_buf_grow(p_buf, p_size) != ROK)
    return RIOPINTFAIL;
  
  ret = read(p_fd, p_buf_read_pos_modifiable(p_buf, used, p_size), p_size);
  
  if (ret >= 0) {
    /* We wrote (read) something - so set the used size to what we used. */
    p_buf_truncate(p_buf, used + ret);
  } else {
    /* We didn't write (read anything) - revert the used size. */
    p_buf_truncate(p_buf, used);
  }
  
  return ret;
}

ssize_t p_io_read_all(p_arr *p_buf, int p_fd, size_t *p_size)
{
  unsigned char *buf;
  ssize_t left;
  ssize_t ret;
  
  /* left = amount left to read. */
  left = *p_size;
  /* *p_size = amount read. */
  *p_size = 0;

  if ((buf = p_buf_read_pos_modifiable(p_buf, *p_size, left)) == NULL)
    return -2;

  while (1) {
    ret = read(p_fd, buf, left);
    if (ret < 0) /* Error reading. */
      return -1;
    
    *p_size += ret;
    left -= ret;

    if (left == 0) /* Read everything - cool! */
      return 1;
     
    if (ret == 0) /* Read all we're going to. */
      return 0;
    
    buf += ret;
  } /* END: while (1) */
}
    

ssize_t p_io_write(int p_fd, p_arr *p_buf, size_t p_size)
{
  const void *buf;
  
  if ((buf = p_buf_read_pos(p_buf, 0, p_size)) == NULL)
    return -2;
  
  return write(p_fd, buf, p_size);
}

ssize_t p_io_write_all(int p_fd, p_arr *p_buf, size_t *p_size)
{
  ssize_t wrote;
  ssize_t ret;
  const unsigned char *buf;

  P_PRIVATE_IOUNIXNET_WRITEALLBEG(p_fd, p_buf, wrote, p_size);
  ret = write(p_fd, buf, *p_size);
  P_PRIVATE_IOUNIXNET_WRITEALLEND(p_fd, p_buf, wrote, p_size);
}
