/*
 * Network API. Uses buffer API. In namespace p_net_.
 *
 * Copyright (c) 2005, Adam Gleave
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the 
 *     documentation and/or other materials provided with the distribution.
 *  3. Neither the name of the copyright holders nor the names of its
 *     contributors may be used to endorse or promote products derived from
 *     this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT OWNER AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

/*
 * @UNSAFE: Whole File
 */

#include <stdio.h>
#include <sys/types.h>
#include <sys/socket.h>

#include "buffer.h"
#include "bufdang.h"
#include "net.h"
#include "priviounixnet.h"
#include "privfail.h"

#define ROK P_RET_OK

#define RNETINTFAIL P_NET_RET_INTFAIL
#define RNETEXTFAIL P_NET_RET_EXTFAIL
#define RNETPARTSENT P_NET_RET_PARTSENT
#define RNETSENT P_NET_RET_SENT


ssize_t p_net_send(int p_sock, p_arr *p_buf, int p_flags, size_t p_size)
{
  const void *buf;
  
  if ((buf = p_buf_read_pos(p_buf, 0, p_size)) == NULL)
    return RNETINTFAIL;
  
  return send(p_sock, buf, p_size, p_flags);
}

int p_net_send_all(int p_sock, p_arr *p_buf, int p_flags, size_t *p_size)
{
  ssize_t ret;
  size_t sent;
  const unsigned char *buf;
  
  P_PRIVATE_IOUNIXNET_WRITEALLBEG(p_sock, p_buf, sent, p_size);
  ret = send(p_sock, buf, *p_size, p_flags);
  P_PRIVATE_IOUNIXNET_WRITEALLEND(p_sock, p_buf, sent, p_size);
}

ssize_t p_net_sendto(int p_sock, p_arr *p_buf, const struct sockaddr *p_to,
                     socklen_t p_tolen, int p_flags, size_t p_size)
{
  const void *buf;
  
  if ((buf = p_buf_read_pos(p_buf, 0, p_size)) == NULL)
    return RNETINTFAIL;
  
  return sendto(p_sock, buf, p_size, p_flags, p_to, p_tolen);
}

int p_net_sendto_all(int p_sock, p_arr *p_buf, const struct sockaddr *p_to,
                     socklen_t p_tolen, int p_flags, size_t *p_size)
{
  ssize_t ret;
  size_t sent;
  const unsigned char *buf;

  P_PRIVATE_IOUNIXNET_WRITEALLBEG(p_sock, p_buf, sent, p_size);
  ret = sendto(p_sock, buf, *p_size, p_flags, p_to, p_tolen);
  P_PRIVATE_IOUNIXNET_WRITEALLEND(p_sock, p_buf, sent, p_size);
}

ssize_t p_net_recv(p_arr *p_buf, int p_sock, int p_flags, size_t p_size)
{
  size_t used;
  ssize_t read;
  
  used = p_buf_get_used(p_buf);
  
  if (p_buf_grow(p_buf, p_size) != ROK)
    return RNETINTFAIL;
  
  read = recv(p_sock, p_buf_read_pos_modifiable(p_buf, used, p_size),
  						p_size, p_flags);
  
  if (read >= 0) {
    /* No errors - truncate buffer to used amount. */
    p_buf_truncate(p_buf, used + read);
	} else {
    /* AIEE! An error. */
    p_buf_truncate(p_buf, used);
  }
  
  return read;
}

ssize_t p_net_recvfrom(p_arr *p_buf, int p_sock, struct sockaddr *p_from, 
                       socklen_t *p_fromlen, int p_flags, size_t p_size)
{
  size_t used;
  ssize_t read;
  
  used = p_buf_get_used(p_buf);
  
  if (p_buf_grow(p_buf, p_size) != ROK)
    return RNETINTFAIL;
  
  read = recvfrom(p_sock, p_buf_read_pos_modifiable(p_buf, used, p_size),
                  p_size, p_flags, p_from, p_fromlen);
  
  if (read >= 0) {
    /* No errors - truncate buffer to used amount. */
    p_buf_truncate(p_buf, used + read);
  } else {
    /* AIEE! An error. */
    p_buf_truncate(p_buf, used);
  }
  
  return read;
}
